from requests_cache import CachedSession
from platformdirs import user_cache_dir
import ipaddress
from collections import defaultdict
from asn_check.rfc6890 import assigned_ranges, assigned_names
import logging

log = logging.getLogger("__main__")

ASN_ROUTES_URL_V4 = 'https://thyme.apnic.net/current/data-raw-table'
ASN_ROUTES_URL_V6 = 'https://thyme.apnic.net/current/ipv6-raw-table'
ASN_NAMES_URL = 'https://ftp.ripe.net/ripe/asnames/asn.txt'


def get_data(
    asn_routes_url_v4: str = ASN_ROUTES_URL_V4,
    asn_routes_url_v6: str = ASN_ROUTES_URL_V6,
    asn_names_url: str = ASN_NAMES_URL,
):
    session = CachedSession(user_cache_dir('asn_check', 'mh'), cache_control=True, backend='filesystem')
    log.info(f"Getting ASN routes v4 from {asn_routes_url_v4}")
    asn_routes_v4 = session.get(asn_routes_url_v4)
    log.info(f"Getting ASN routes v6 from {asn_routes_url_v6}")
    asn_routes_v6 = session.get(asn_routes_url_v6)
    log.info(f"Getting ASN names from {asn_names_url}")
    asn_names = session.get(asn_names_url)
    return asn_routes_v4.text, asn_routes_v6.text, asn_names.text


def parse_asn_routes(asn_routes_v4: str, asn_routes_v6: str):
    v4 = parse_asn_routes_v4(asn_routes_v4)
    v6 = parse_asn_routes_v6(asn_routes_v6)
    total_dict = dict()
    for key in set(list(v4.keys()) + list(v6.keys())):
        total_dict[key] = list(set(v4.get(key, []) + v6.get(key, [])))
    for key, value in assigned_ranges.items():
        total_dict[key] = value
    return total_dict


def parse_asn_routes_v4(asn_routes: str):
    """
    -> dict[str,list[ipaddress.IPv4Network]]
    """
    result = defaultdict(list)
    for line in asn_routes.splitlines():
        try:
            p = line.split('\t')
            result[p[1]].append(ipaddress.IPv4Network(p[0]))
        except:
            log.error(f"Invalid ASN route format: '{line}'")
    return result


def parse_asn_routes_v6(asn_routes: str):
    """
    -> dict[str,list[ipaddress.IPv6Network]]
    """
    result = defaultdict(list)
    for line in asn_routes.splitlines():
        try:
            p = line.split(' ')  # for some reason, the v6 uses spaces, not tabs
            result[p[-1]].append(ipaddress.IPv6Network(p[0]))
        except:
            log.error(f"Invalid ASN route format: '{line}'")
    return result


def parse_asn_names(asn_names: str):
    """
    -> dict[str,dict[str,str]]
    """
    result = dict()
    for line in asn_names.splitlines():
        try:
            sp = line.find(' ')
            c = line.find(',')
            asn = line[:sp]
            name = line[sp + 1 : c]
            code = line[c + 2 :]
            result[asn] = {'name': name, 'country_code': code}
        except:
            log.error(f"Invalid ASN name format: '{line}'")
    for key, value in assigned_names.items():
        result[key] = value
    return result



import pytest
import ipaddress

from asn_check import data


def test_parse_asn_routes_v4():
    cases = [
        {
            "label": "happy",  # I don't really expect them to break the format
            "input": {
                "asn_routes": """1.0.0.0/24	13335
1.0.4.0/22	38803
1.0.5.0/24	38803
1.0.16.0/24	2519
1.0.128.0/17	23969
1.0.128.0/18	23969
1.0.128.0/19	23969
1.0.128.0/24	23969
1.0.129.0/24	23969
1.0.130.0/23	23969"""
            },
            "output": {
                "13335": {ipaddress.IPv4Network("1.0.0.0/24")},
                "38803": {ipaddress.IPv4Network("1.0.4.0/22"), ipaddress.IPv4Network("1.0.5.0/24")},
                "2519": {ipaddress.IPv4Network("1.0.16.0/24")},
                "23969": {
                    ipaddress.IPv4Network("1.0.128.0/17"),
                    ipaddress.IPv4Network("1.0.128.0/18"),
                    ipaddress.IPv4Network("1.0.128.0/19"),
                    ipaddress.IPv4Network("1.0.128.0/24"),
                    ipaddress.IPv4Network("1.0.129.0/24"),
                    ipaddress.IPv4Network("1.0.130.0/23"),
                },
            },
            "ex": None,
        },
    ]
    for case in cases:
        try:
            output = data.parse_asn_routes_v4(**case["input"])
            assert {k: set(v) for k, v in output.items()} == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))


def test_parse_asn_routes_v6():
    cases = [
        {
            "label": "happy",  # I don't really expect them to break the format
            "input": {
                "asn_routes": """2a03:1980:d1ff::/48     48260
2a03:1980:d200::/40     48260
2a03:1980:d400::/40     48260
2a03:1980:d4ff::/48     48260
2a03:1981::/32          48260
2a03:1984::/32          48260
2a03:19c0::/32         204141
2a03:1a20::/48         207467
2a03:1a20:10::/48      207467
2a03:1a60::/32         197161"""
            },
            "output": {
                "204141": {ipaddress.IPv6Network("2a03:19c0::/32")},
                "207467": {ipaddress.IPv6Network("2a03:1a20::/48"), ipaddress.IPv6Network("2a03:1a20:10::/48")},
                "197161": {ipaddress.IPv6Network("2a03:1a60::/32")},
                "48260": {
                    ipaddress.IPv6Network("2a03:1984::/32"),
                    ipaddress.IPv6Network("2a03:1981::/32"),
                    ipaddress.IPv6Network("2a03:1980:d1ff::/48"),
                    ipaddress.IPv6Network("2a03:1980:d4ff::/48"),
                    ipaddress.IPv6Network("2a03:1980:d200::/40"),
                    ipaddress.IPv6Network("2a03:1980:d400::/40"),
                },
            },
            "ex": None,
        },
    ]
    for case in cases:
        try:
            output = data.parse_asn_routes_v6(**case["input"])
            assert {k: set(v) for k, v in output.items()} == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))


def test_parse_asn_names():
    cases = [
        {
            "label": "happy",  # I don't really expect them to break the format
            "input": {
                "asn_names": """25 UCB, US
26 CORNELL, US
27 UMDNET, US
28 DFVLR-SYS Deutsches Zentrum fuer Luft- und Raumfahrt e.V., DE"""
            },
            "output": {
                "25": {"name": "UCB", "country_code": "US"},
                "26": {"name": "CORNELL", "country_code": "US"},
                "27": {"name": "UMDNET", "country_code": "US"},
                "28": {"name": "DFVLR-SYS Deutsches Zentrum fuer Luft- und Raumfahrt e.V.", "country_code": "DE"},
                "rfc6890-t1": {"name": "This host on this network", "country_code": "IANA"},
                "rfc6890-t2": {"name": "Private-Use Networks", "country_code": "IANA"},
                "rfc6890-t3": {"name": "Shared Address Space", "country_code": "IANA"},
                "rfc6890-t4": {"name": "Loopback", "country_code": "IANA"},
                "rfc6890-t5": {"name": "Link Local", "country_code": "IANA"},
                "rfc6890-t6": {"name": "Private-Use Networks", "country_code": "IANA"},
                "rfc6890-t7": {"name": "IETF Protocol Assignments", "country_code": "IANA"},
                "rfc6890-t8": {"name": "DS-Lite", "country_code": "IANA"},
                "rfc6890-t9": {"name": "TEST-NET-1", "country_code": "IANA"},
                "rfc6890-t10": {"name": "6to4 Relay Anycast", "country_code": "IANA"},
                "rfc6890-t11": {"name": "Private-Use Networks", "country_code": "IANA"},
                "rfc6890-t12": {"name": "Network Interconnect Device Benchmark Testing", "country_code": "IANA"},
                "rfc6890-t13": {"name": "TEST-NET-2", "country_code": "IANA"},
                "rfc6890-t14": {"name": "TEST-NET-3", "country_code": "IANA"},
                "rfc6890-t15": {"name": "Reserved for Future Use", "country_code": "IANA"},
                "rfc6890-t16": {"name": "Limited Broadcast", "country_code": "IANA"},
                "rfc6890-t17": {"name": "Loopback", "country_code": "IANA"},
                "rfc6890-t18": {"name": "Unspecified Address", "country_code": "IANA"},
                "rfc6890-t19": {"name": "IPv4-IPv6 Translation Address", "country_code": "IANA"},
                "rfc6890-t20": {"name": "IPv4-Mapped Address", "country_code": "IANA"},
                "rfc6890-t21": {"name": "Discard-Only Address Block", "country_code": "IANA"},
                "rfc6890-t22": {"name": "IETF Protocol Assignments", "country_code": "IANA"},
                "rfc6890-t23": {"name": "TEREDO", "country_code": "IANA"},
                "rfc6890-t24": {"name": "Benchmarking", "country_code": "IANA"},
                "rfc6890-t25": {"name": "Documentation", "country_code": "IANA"},
                "rfc6890-t26": {"name": "ORCHID", "country_code": "IANA"},
                "rfc6890-t27": {"name": "6to4", "country_code": "IANA"},
                "rfc6890-t28": {"name": "Unique-Local", "country_code": "IANA"},
                "rfc6890-t29": {"name": "Linked-Scoped Unicast", "country_code": "IANA"},
            },
            "ex": None,
        },
    ]
    for case in cases:
        try:
            output = data.parse_asn_names(**case["input"])
            assert output == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))



import ipaddress
import logging

log = logging.getLogger("__main__")

# https://www.rfc-editor.org/rfc/rfc6890.txt
assigned_ranges = {
    "rfc6890-t1": [ipaddress.IPv4Network("0.0.0.0/8")],
    "rfc6890-t2": [ipaddress.IPv4Network("10.0.0.0/8")],
    "rfc6890-t3": [ipaddress.IPv4Network("100.64.0.0/10")],
    "rfc6890-t4": [ipaddress.IPv4Network("127.0.0.0/8")],
    "rfc6890-t5": [ipaddress.IPv4Network("169.254.0.0/16")],
    "rfc6890-t6": [ipaddress.IPv4Network("172.16.0.0/12")],
    "rfc6890-t7": [ipaddress.IPv4Network("192.0.0.0/24")],  # this feels wrong, but I don't care
    "rfc6890-t8": [ipaddress.IPv4Network("192.0.0.0/29")],  # this feels wrong, but I don't care
    "rfc6890-t9": [ipaddress.IPv4Network("192.0.2.0/24")],
    "rfc6890-t10": [ipaddress.IPv4Network("192.88.99.0/24")],
    "rfc6890-t11": [ipaddress.IPv4Network("192.168.0.0/16")],
    "rfc6890-t12": [ipaddress.IPv4Network("198.18.0.0/15")],
    "rfc6890-t13": [ipaddress.IPv4Network("198.51.100.0/24")],
    "rfc6890-t14": [ipaddress.IPv4Network("203.0.113.0/24")],
    "rfc6890-t15": [ipaddress.IPv4Network("240.0.0.0/4")],
    "rfc6890-t16": [ipaddress.IPv4Network("255.255.255.255/32")],
    "rfc6890-t17": [ipaddress.IPv6Network("::1/128")],
    "rfc6890-t18": [ipaddress.IPv6Network("::/128")],
    "rfc6890-t19": [ipaddress.IPv6Network("64:ff9b::/96")],
    "rfc6890-t20": [ipaddress.IPv6Network("::ffff:0:0/96")],
    "rfc6890-t21": [ipaddress.IPv6Network("100::/64")],
    "rfc6890-t22": [ipaddress.IPv6Network("2001::/23")],
    "rfc6890-t23": [ipaddress.IPv6Network("2001::/32")],
    "rfc6890-t24": [ipaddress.IPv6Network("2001:2::/48")],
    "rfc6890-t25": [ipaddress.IPv6Network("2001:db8::/32")],
    "rfc6890-t26": [ipaddress.IPv6Network("2001:10::/28")],
    "rfc6890-t27": [ipaddress.IPv6Network("2002::/16")],
    "rfc6890-t28": [ipaddress.IPv6Network("fc00::/7")],
    "rfc6890-t29": [ipaddress.IPv6Network("fe80::/10")],
}

assigned_names = {
    "rfc6890-t1": {"name": "This host on this network", "country_code": "IANA"},
    "rfc6890-t2": {"name": "Private-Use Networks", "country_code": "IANA"},
    "rfc6890-t3": {"name": "Shared Address Space", "country_code": "IANA"},
    "rfc6890-t4": {"name": "Loopback", "country_code": "IANA"},
    "rfc6890-t5": {"name": "Link Local", "country_code": "IANA"},
    "rfc6890-t6": {"name": "Private-Use Networks", "country_code": "IANA"},
    "rfc6890-t7": {"name": "IETF Protocol Assignments", "country_code": "IANA"},
    "rfc6890-t8": {"name": "DS-Lite", "country_code": "IANA"},
    "rfc6890-t9": {"name": "TEST-NET-1", "country_code": "IANA"},
    "rfc6890-t10": {"name": "6to4 Relay Anycast", "country_code": "IANA"},
    "rfc6890-t11": {"name": "Private-Use Networks", "country_code": "IANA"},
    "rfc6890-t12": {"name": "Network Interconnect Device Benchmark Testing", "country_code": "IANA"},
    "rfc6890-t13": {"name": "TEST-NET-2", "country_code": "IANA"},
    "rfc6890-t14": {"name": "TEST-NET-3", "country_code": "IANA"},
    "rfc6890-t15": {"name": "Reserved for Future Use", "country_code": "IANA"},
    "rfc6890-t16": {"name": "Limited Broadcast", "country_code": "IANA"},
    "rfc6890-t17": {"name": "Loopback", "country_code": "IANA"},
    "rfc6890-t18": {"name": "Unspecified Address", "country_code": "IANA"},
    "rfc6890-t19": {"name": "IPv4-IPv6 Translation Address", "country_code": "IANA"},
    "rfc6890-t20": {"name": "IPv4-Mapped Address", "country_code": "IANA"},
    "rfc6890-t21": {"name": "Discard-Only Address Block", "country_code": "IANA"},
    "rfc6890-t22": {"name": "IETF Protocol Assignments", "country_code": "IANA"},
    "rfc6890-t23": {"name": "TEREDO", "country_code": "IANA"},
    "rfc6890-t24": {"name": "Benchmarking", "country_code": "IANA"},
    "rfc6890-t25": {"name": "Documentation", "country_code": "IANA"},
    "rfc6890-t26": {"name": "ORCHID", "country_code": "IANA"},
    "rfc6890-t27": {"name": "6to4", "country_code": "IANA"},
    "rfc6890-t28": {"name": "Unique-Local", "country_code": "IANA"},
    "rfc6890-t29": {"name": "Linked-Scoped Unicast", "country_code": "IANA"},
}



import ipaddress
from typing import Union


class IPNode:
    def __init__(self, value=None):
        self.one = None
        self.zero = None
        self.value = value


class IPTree:
    def __init__(self):
        self.root_v4 = IPNode()
        self.root_v6 = IPNode()

    def add_ip(self, ip: Union[ipaddress.IPv4Network, ipaddress.IPv6Network], label: str):
        if type(ip) == ipaddress.IPv4Network:
            self.add_ipv4(ip, label)
        elif type(ip) == ipaddress.IPv6Network:
            self.add_ipv6(ip, label)
        else:
            raise ValueError(f"Unsupported IP type {type(ip)} (supported ipaddress.IPv4Network, ipaddress.IPv6Network)")

    def search(self, ip: Union[ipaddress.IPv4Address, ipaddress.IPv6Address]):
        if type(ip) == ipaddress.IPv4Address:
            return self.search_ipv4(ip)
        elif type(ip) == ipaddress.IPv6Address:
            return self.search_ipv6(ip)
        else:
            raise ValueError(f"Unsupported IP type {type(ip)} (supported ipaddress.IPv4Address, ipaddress.IPv6Address)")

    def add_ipv4(self, ip: ipaddress.IPv4Network, label: str):
        addr_ip_netmask = int(ip.netmask)
        addr_ip = int(ip.network_address)
        node_pointer = self.root_v4
        for i in range(32):
            if not addr_ip_netmask & 1 << 31 - i:
                node_pointer.value = label
                break
            if addr_ip & 1 << 31 - i:
                if node_pointer.one is None:
                    node_pointer.one = IPNode()
                node_pointer = node_pointer.one
            else:
                if node_pointer.zero is None:
                    node_pointer.zero = IPNode()
                node_pointer = node_pointer.zero
        node_pointer.value = label

    def search_ipv4(self, ip: ipaddress.IPv4Address) -> str:
        node_pointer = self.root_v4
        addr_ip = int(ip)
        for i in range(32):
            if node_pointer is None:
                return None
            if node_pointer.value is not None:
                return node_pointer.value
            if addr_ip & 1 << 31 - i:
                node_pointer = node_pointer.one
            else:
                node_pointer = node_pointer.zero
        if node_pointer is not None and node_pointer.value is not None:
            return node_pointer.value
        return None

    def add_ipv6(self, ip: ipaddress.IPv6Network, label: str):
        addr_ip_netmask = int(ip.netmask)
        addr_ip = int(ip.network_address)
        node_pointer = self.root_v6
        for i in range(128):
            if not addr_ip_netmask & 1 << 127 - i:
                node_pointer.value = label
                break
            if addr_ip & 1 << 127 - i:
                if node_pointer.one is None:
                    node_pointer.one = IPNode()
                node_pointer = node_pointer.one
            else:
                if node_pointer.zero is None:
                    node_pointer.zero = IPNode()
                node_pointer = node_pointer.zero
        node_pointer.value = label

    def search_ipv6(self, ip: ipaddress.IPv6Address) -> str:
        node_pointer = self.root_v6
        addr_ip = int(ip)
        for i in range(128):
            if node_pointer is None:
                return None
            if node_pointer.value is not None:
                return node_pointer.value
            if addr_ip & 1 << 127 - i:
                node_pointer = node_pointer.one
            else:
                node_pointer = node_pointer.zero
        if node_pointer is not None and node_pointer.value is not None:
            return node_pointer.value
        return None
